/*
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *   * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 */

package org.wso2.registry;

import java.util.Date;

/**
 * API for the Registry. Java applications should use this API to access the registry.
 */
public interface Registry {

    /**
     * Returns the resource in the given path.
     *
     * @param path Path of the resource. e.g. /project1/server/deployment.xml
     * @return Resource instance
     * @throws RegistryException is thrown if the resource is not in the
     *                           registry
     */
    Resource get(String path) throws RegistryException;

    /**
     * Check whether a resource exist in the given path
     *
     * @param path Path of the resource to be checked
     * @return true if a resource exist in the given path, false otherwise.
     * @throws RegistryException if an error occurs
     */
    public boolean resourceExists(String path) throws RegistryException;

    /**
     * Adds or updates resources in the registry. If there is no resource in the given path,
     * resource is added. If a resource already exist in the given path, it will be replaced with
     * the new resource.
     *
     * @param path   Path of the new resource
     * @param resource Resource instance for the new resource
     * @throws org.wso2.registry.RegistryException is thrown if the storage exception is encountered.
     */
    void put(String path, Resource resource) throws RegistryException;

    /**
     * Deletes the resource in the given path. If the path refers to a directory, all child
     * resources of the directory will also be deleted.
     *
     * @param path Path of the resource to be deleted.
     * @throws org.wso2.registry.RegistryException is thrown depending on the implementation.
     */
    void delete(String path) throws RegistryException;

    /**
     * Returns the paths of the versions of the resource located at the given path.
     * Version paths are returned in the form /projects/myresource?v=12
     *
     * @param path Path of a current version of a resource
     * @return Paths of versions
     * @throws RegistryException if there is an error
     */
    String[] getVersions(String path) throws RegistryException;

    /**
     * Reverts a resource to a given version.
     *
     * @param versionPath path of the version to be reverted. It is not necessary to provide the
     * path of the resource as it can be derived from the version path.
     * 
     * @throws RegistryException if there is an error
     */
    void restoreVersion(String versionPath) throws RegistryException;

    ////////////////////////////////////////////////////////
    // Tagging
    ////////////////////////////////////////////////////////

    /**
     * Applies the given tag to the resource in the given path. If the given tag is not defined
     * in the registry, it will be defined.
     *
     * @param resourcePath Path of the resource to be tagged.
     * @param tag Tag. Any string can be used for the tag.
     * @throws RegistryException is thrown if a resource does not exist in the given path.
     */
    void applyTag(String resourcePath, String tag) throws RegistryException;

    /**
     * Returns the paths of the resources, which are tagged using the given tag.
     *
     * @param tag Tag
     * @return Resource paths
     * @throws RegistryException if an error occurs
     */
    TaggedResourcePath[] getResourcePathsWithTag(String tag) throws RegistryException;

    /**
     * Returns all tags used for tagging the given resource.
     *
     * @param resourcePath Path of the resource
     * @return Tags tag names
     * @throws RegistryException is thrown if a resource does not exist in the given path.
     */
    Tag[] getTags(String resourcePath) throws RegistryException;

    /**
     * Removes a tag on a resource. If the resource at the path is owned by the current user, all
     * taggings done using the given tag will be removed. If the resource is not owned by the
     * current user, only the tagging done by the current user will be removed.
     *
     * @param path Resource path tagged with the given tag.
     * @param tag Name of the tag to be removed.
     * @throws RegistryException if there's a problem
     */
    public void removeTag(String path, String tag) throws RegistryException;

    ////////////////////////////////////////////////////////
    // Comments
    ////////////////////////////////////////////////////////

    /**
     * Adds a comment to a resource.
     *
     * @param resourcePath Path of the resource to add the comment.
     * @param comment Comment instance for the new comment.
     * @throws RegistryException is thrown if a resource does not exist in the given path.
     */
    void addComment(String resourcePath, Comment comment) throws RegistryException;

    /**
     * Retruns all comments for the given resource.
     *
     * @param resourcePath Path of the resource.
     * @return Array of Comment objects.
     * @throws RegistryException if an error occurs
     */
    Comment[] getComments(String resourcePath) throws RegistryException;

    ////////////////////////////////////////////////////////
    // Ratings
    ////////////////////////////////////////////////////////

    /**
     * Rates the given resource.
     *
     * @param resourcePath Path of the resource.
     * @param rating rating value between 1 and 5.
     * @throws RegistryException is thrown if a resource does not exist in the given path.
     */
    void rateResource(String resourcePath, int rating) throws RegistryException;

    /**
     * Returns the avarage rating for the given resource. This is the average of all ratings done by
     * all users for the given resource.
     *
     * @param resourcePath Path of the resource.
     * @return Average rating between 1 and 5.
     * @throws RegistryException if an error occurs
     */
    float getAverageRating(String resourcePath) throws RegistryException;

    /**
     * Returns the rating given to the specified resource by the given user
     *
     * @param path Path of the resource
     * @param userName username of the user
     * @return rating given by the given user
     * @throws RegistryException if there is a problem
     */
    int getRating(String path, String userName) throws RegistryException;

    /**
     * Executes a custom query defined by users. Returned RowSet contains fields of tables specified
     * by the custom query.
     *
     * @param path Path of the query to execute.
     * @param parameters Query parameters.
     * @return RowSet containing the fields specified in the query.
     * @throws RegistryException depends on the implementation.
     */
    Resource executeQuery (String path, Object[] parameters) throws RegistryException;

    /**
     * Returns the logs of the activities occured in the registy.
     *
     * @param resourcePath If given, only the logs related to the resource path will be returned.
     * If null, logs for all resources will be returned.
     *
     * @param action Only the logs pertaining to this action will be returned.
     *
     * @param userName If given, only the logs for activities done by the given user will be
     * returned. If null, logs for all users will be returned.
     *
     * @param from If given, logs for activities occured after the given date will be returned. If
     * null, there will not be a bound for the starting date.
     *
     * @param to If given, logs for activities occured before the given date will be returned. If
     * null, there will not be a bound for the ending date.
     *
     * @param descending If true, returned activities will be in the descending order. If false,
     * returned activities will be in the ascending order.
     *
     * @return Array of LogEntry objects representing the logs
     *
     * @throws RegistryException if there is a problem
     */
    LogEntry[] getLogs(String resourcePath, int action, String userName, Date from, Date to, boolean descending)
            throws RegistryException;
}
