/*
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *   * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 */

package org.wso2.registry.secure;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.registry.*;
import org.wso2.usermanager.Authenticator;
import org.wso2.usermanager.Authorizer;
import org.wso2.usermanager.Realm;
import org.wso2.usermanager.UserManagerException;
import org.wso2.usermanager.acl.realm.AuthorizingRealm;
import org.wso2.usermanager.acl.realm.AuthorizingRealmConfig;

import java.util.Date;

/**
 * Registry implementation to perform authentication and authorization. Each user should use a
 * separate instance of the SecureRegistry initialized with the user name and a back-end Registry.
 * All SecureRegistry instances should be initialized using the same back-end registry
 * implementation. SecureRegistry verifies whether the user has necessary permissions to perform
 * the requested action and delegates the request to the back-end registry if the authorization
 * succeeds. It also sets user name as a thread local variable before delegating the request, so
 * that the back-end registry can use that user name to associate user information with performed
 * operation (e.g. who commented on the resource /projects/abcbank/orders.doc).
 *
 * Resource path is used as the unique ID of the resource to associate permissions. So when an
 * operation is request on a certain path, we can first check the permissions for that path, without
 * even touching the actual resource. This has a side effect that if a user has requested an
 * operation on a non-existing resource, he gets an unauthorized operation message rather than a
 * resource does not exist message. This is good as we don't want unauthorized users to check if
 * a resource exist in a particular path.
 */
public class SecureRegistry implements Registry {

    private static final Log log = LogFactory.getLog(SecureRegistry.class);

    private String userID = RegistryConstants.ANONYMOUS_USER;
    private Registry registry;
    private Authorizer authorizer;
    private RegistryUserManager userManager;

    public SecureRegistry(String userID, String password, Registry registry, Realm defaultRealm)
            throws RegistryException {

        Realm userRealm;        
        try {

            userRealm = new AuthorizingRealm();
            AuthorizingRealmConfig config = (AuthorizingRealmConfig) userRealm
                    .getRealmConfiguration();
            config.setRealm(defaultRealm);
            config.setAuthenticatedUserName(userID);
            userRealm.init(config);
            
            Authenticator authenticator = userRealm.getAuthenticator();
            authorizer = userRealm.getAuthorizer();

            if (!authenticator.authenticate(userID, password)) {
                String msg = "Attempted to authenticate invalid user.";
                log.info(msg);
                throw new RegistryException(msg);
            }
        } catch (UserManagerException e) {
            String msg = "Could not check authentication. \nCaused by " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg);
        }

        this.userID = userID;
        this.registry = registry;
        this.userManager = new RegistryUserManager(userRealm);
    }

    /**
     * Non-authenticating contructor. Use this constructor if the user is already authenticated
     * outside the secure registry.
     *
     * @param userID user name of the user of the secure registry.
     * @param registry core registry to use by the secure registry.
     * @param defaultRealm default (non-authorizing) realm.
     * @throws RegistryException
     */
    public SecureRegistry(String userID, Registry registry, Realm defaultRealm)
            throws RegistryException {

        Realm userRealm;
        try {

            userRealm = new AuthorizingRealm();
            AuthorizingRealmConfig config = (AuthorizingRealmConfig) userRealm
                    .getRealmConfiguration();
            config.setRealm(defaultRealm);
            config.setAuthenticatedUserName(userID);
            userRealm.init(config);

            authorizer = userRealm.getAuthorizer();

        } catch (UserManagerException e) {
            String msg = "Could not check authentication. \nCaused by " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg);
        }

        this.userID = userID;
        this.registry = registry;
        this.userManager = new RegistryUserManager(userRealm);
    }

    public String getUserID() {
        return userID;
    }

    public RegistryUserManager getUserManager() {
        return userManager;
    }

    public Resource get(String path) throws RegistryException {

        // if the user has permission to the current version, he will get permission to all
        // previous versions of the same resource
        String authorizationPath = path;
        if (path.indexOf("?") > 0) {
            authorizationPath = path.split("\\?")[0];
        }

        try {
            if (!authorizer.isUserAuthorized(userID, authorizationPath, ActionConstants.GET)) {
                String msg = "Attempted to perform unauthorized operation (" + ActionConstants.GET +
                             ") on " + authorizationPath + " by " + userID;
                log.info(msg);
                throw new AuthorizationFailedException(msg);
            }
        } catch (UserManagerException e) {
            String msg = "Could not check authorization. \nCaused by " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg);
        }

        User.setCurrentUser(userID);
        return registry.get(path);
    }

    public boolean resourceExists(String path) throws RegistryException {
        User.setCurrentUser(userID);
        return registry.resourceExists(path);
    }

    /**
     * Checks if the user has "PUT" permissions to the existing ascendent path of the given path.
     * e.g. if a user wants to put a resource in to the path "/d1/d2/d3/r1" and only "/d1/d2"
     * currently exist in the registry, that user should have PUT permission to "/d1/d2".
     *
     * If the user has PUT permission, delegates the request to the underlying jdbc registry.
     *
     * Then sets the default permissions for newly created paths. i.e. in the above sample, default
     * permissions are set for "/d1/d2/d3" and "/d1/d2/d3/r1".
     *
     * @param path Path to put the new resource
     * @param resource New resource
     * @throws RegistryException
     */
    public void put(String path, Resource resource) throws RegistryException {

        // find the existing ascendent path of the given path
        String pathToSplit = path;
        if (pathToSplit.startsWith(RegistryConstants.PATH_SEPARATOR)) {
            pathToSplit = pathToSplit.substring(RegistryConstants.PATH_SEPARATOR.length());
        }

        if (pathToSplit.endsWith(RegistryConstants.PATH_SEPARATOR)) {
            pathToSplit = pathToSplit.substring(
                    0, pathToSplit.length() - RegistryConstants.PATH_SEPARATOR.length());
        }

        String[] parts = pathToSplit.split(RegistryConstants.PATH_SEPARATOR);
        int partCount = parts.length;
        String existingAscendant;
        do {
            existingAscendant = "";
            if (partCount == 0) {
                existingAscendant = RegistryConstants.ROOT_PATH;
            }

            for (int i = 0; i < partCount; i++) {
                existingAscendant = existingAscendant + RegistryConstants.PATH_SEPARATOR + parts[i];
            }
            partCount--;
        } while (!registry.resourceExists(existingAscendant));

        // check if the user has put permission for that path
        try {
            if (!authorizer.isUserAuthorized(userID, existingAscendant, ActionConstants.PUT)) {
                String msg = "Attempted to perform unauthorized operation.";
                log.info(msg);
                throw new AuthorizationFailedException(msg);
            }
        } catch (UserManagerException e) {
            String msg = "Could not check authorization. \nCaused by " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg);
        }

        User.setCurrentUser(userID);

//        boolean updatedOnly = false;
//        if (registry.resourceExists(path)) {
//            updatedOnly = true;
//        }

        registry.put(path, resource);

        // we should set default permissions for all newly created paths
        //if (!updatedOnly) {
        //    String newPath = RegistryConstants.ROOT_PATH.equals(existingAscendant)?
        //            "": existingAscendant;
        //    for (int i = partCount + 1; i < parts.length; i++) {
        //        newPath = newPath + RegistryConstants.PATH_SEPARATOR + parts[i];
        //        setDefaultAuthorizations(newPath, userID);
        //    }
        //}
    }

    public void delete(String path) throws RegistryException {

        try {
            if (!authorizer.isUserAuthorized(userID, path, ActionConstants.DELETE)) {
                String msg = "Attempted to perform unauthorized operation.";
                log.info(msg);
                throw new AuthorizationFailedException(msg);
            }
        } catch (UserManagerException e) {
            String msg = "Could not check authorization. \nCaused by " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg);
        }

        User.setCurrentUser(userID);
        registry.delete(path);
    }


    public String[] getVersions(String path) throws RegistryException {

        // TODO: check authorizations

        User.setCurrentUser(userID);
        return registry.getVersions(path);
    }

    public void restoreVersion(String versionPath) throws RegistryException {

        // TODO: check authorizations

        User.setCurrentUser(userID);
        registry.restoreVersion(versionPath);
    }

    public void applyTag(String resourcePath, String tag) throws RegistryException {
        User.setCurrentUser(userID);
        registry.applyTag(resourcePath, tag);
    }

    public TaggedResourcePath[] getResourcePathsWithTag(String tag) throws RegistryException {
        User.setCurrentUser(userID);
        return registry.getResourcePathsWithTag(tag);
    }

    public Tag[] getTags(String resourcePath) throws RegistryException {
        User.setCurrentUser(userID);
        return registry.getTags(resourcePath);
    }

    public void removeTag(String path, String tag) throws RegistryException {
        User.setCurrentUser(userID);
        registry.removeTag(path, tag);
    }

    public void addComment(String resourcePath, Comment comment) throws RegistryException {
        User.setCurrentUser(userID);
        registry.addComment(resourcePath, comment);
    }

    public Comment[] getComments(String resourcePath) throws RegistryException {
        User.setCurrentUser(userID);
        return registry.getComments(resourcePath);
    }

    public void rateResource(String resourcePath, int rating) throws RegistryException {
        User.setCurrentUser(userID);
        registry.rateResource(resourcePath, rating);
    }

    public float getAverageRating(String resourcePath) throws RegistryException {
        User.setCurrentUser(userID);
        return registry.getAverageRating(resourcePath);
    }

    public int getRating(String path, String userName) throws RegistryException {
        User.setCurrentUser(userID);
        return registry.getRating(path, userName);
    }

    public Resource executeQuery(String path, Object[] parameters) throws RegistryException {
        User.setCurrentUser(userID);
        return registry.executeQuery(path, parameters);
    }

    public LogEntry[] getLogs(String resourcePath, int action, String userName, Date from, Date to, boolean descending)
            throws RegistryException {
        User.setCurrentUser(userID);
        return registry.getLogs(resourcePath, action, userName, from, to, descending);
    }
}
